from .layer import *
from .net import Net
from .io import read_net, onnx2planer
from .util import *

# planer array library
pal = None

def core(obj):
	global pal
	pal = obj
	from . import util, layer, net, io
	util.np = layer.np = net.np = io.np = obj
	if 'asnumpy' in dir(pal):
		pal.cpu = pal.asnumpy
	else:
		pal.cpu = pal.asarray
	return pal
	print('\nuser switch engine:', core.__name__)    

def asnumpy(arr, **key): return pal.cpu(arr, **key)

def asarray(arr, **key): return pal.asarray(arr, **key)

try:
    import cupy
    core(cupy)
    print('using cupy engine, gpu powered!')
except:
    import numpy as np
    core(np)
    print('using numpy engine, install cupy would be faster.')


from .util import conv, maxpool, upsample, avgpool, np


class Layer:
    name = 'layer'

    def __init__(self): pass

    def forward(self, x): pass

    def backward(self, grad_y): pass

    def para(self): return None

    def load(self, buf): return 0

    def __call__(self, x):
        return self.forward(x)


class Dense(Layer):
    name = 'dense'

    def __init__(self, c, n):
        self.K = np.zeros((n, c), dtype=np.float32)
        self.bias = np.zeros(n, dtype=np.float32)

    def para(self): return self.K.shape

    def forward(self, x):
        y = x.dot(self.K.T)
        y += self.bias.reshape((1, -1))
        return y

    def load(self, buf):
        sk, sb = self.K.size, self.bias.size
        self.K.ravel()[:] = buf[:sk]
        self.bias.ravel()[:] = buf[sk:sk+sb]

        return sk + sb


class Conv2d(Layer):
    name = 'conv'

    def __init__(self, c, n, w, g=1, s=1, d=1, p=1):
        """
        c: in_channels
        n: out_channels
        w: kernel_size
        g: groups
        s: stride
        d: dilation
        p: padding
        """
        self.n, self.c, self.w = n, c, w
        self.g, self.s, self.d = g, s, d
        self.p = p

        self.K = np.zeros((n, c, w, w), dtype=np.float32)
        self.bias = np.zeros(n, dtype=np.float32)

    def para(self):
        return self.n, self.c, self.w, self.s, self.d, self.p

    def forward(self, x):
        out = conv(x, self.K, self.g, (self.p, self.p), (self.s, self.s), (self.d, self.d))
        out += self.bias.reshape(1, -1, 1, 1)
        return out

    def load(self, buf):
        sk, sb = self.K.size, self.bias.size
        self.K.ravel()[:] = buf[:sk]
        self.bias.ravel()[:] = buf[sk:sk+sb]        
        return sk + sb


class ReLU(Layer):
    name = 'relu'

    def __init__(self): pass

    def forward(self, x):
        #return np.maximum(x, 0, out=x)
        msk = x>0; x *= msk; return x

class LeakyReLU(Layer):
    name = 'leakyrelu'

    def __init__(self, alpha=0.2):
        self.alpha = alpha

    def forward(self, x):
        xalpha = x * self.alpha
        x2 = np.array([x, xalpha])
        x2 = x2.reshape((2,-1)).max(axis=0)
        return x2.reshape(x.shape)

class Flatten(Layer):
    name = 'flatten'

    def __init__(self): pass

    def forward(self, x):
        return x.reshape((x.shape[0], -1))


class Sigmoid(Layer):
    name = 'sigmoid'

    def __init__(self): pass

    def forward(self, x):
        return 1/(1 + np.exp(-x))


class Softmax(Layer):
    name = 'softmax'

    def __init__(self, axis=-1):
        self.axis = axis

    def forward(self, x):
        eX = np.exp((x.T - np.max(x, axis=self.axis)).T)
        return (eX.T / eX.sum(axis=self.axis)).T


class Maxpool(Layer):
    name = 'maxpool'

    def __init__(self, w=(2,2), pad=(0,0), stride=(2,2)):
        self.w, self.pad, self.stride = w, pad, stride

    def para(self): return (self.w, self.pad, self.stride)

    def forward(self, x):
        return maxpool(x, self.w, self.pad, self.stride)


class Avgpool(Layer):
    name = 'avgpool'

    def __init__(self, w=(2,2), pad=(0,0), stride=(2,2)):
        self.w, self.pad, self.stride = w, pad, stride

    def para(self): return (self.w, self.stride)

    def forward(self, x):
        return avgpool(x, self.w, self.pad, self.stride)


class GlobalAveragePool(Layer):
    name = 'gap'
    def __init__(self): pass

    def forward(self, x):
        return x.mean(axis=(-2, -1), keepdims=True)


class UpSample(Layer):
    name = 'upsample'

    def __init__(self, k, mode):
        self.k = k
        self.mode = mode

    def para(self): return (self.k,)

    def forward(self, x):
        return upsample(x, self.k, self.mode)


class Concatenate(Layer):
    name = 'concat'

    def __init__(self, axis): self.axis=axis

    def forward(self, x):
        return np.concatenate(x, axis=self.axis)


class Add(Layer):
    name = 'add'

    def __init__(self): pass

    def forward(self, x):
        x[0] += x[1]
        return x[0]


class Pow(Layer):
    name = 'pow'
    
    def __init__(self, n): self.n = n
    
    def forward(self, x):
        return np.power(x, self.n)

    
class Div(Layer):
    name = 'div'
    
    def __init__(self): pass
    
    def forward(self, x):
        return x[0]/x[1]

class ReduceSum(Layer):
    name = 'reducesum'
    
    def __init__(self, axis, keep_dim):
        self.axis = axis
        self.keep_dim = keep_dim
    
    def forward(self, x):
        if self.keep_dim:
            return np.expand_dims(x.sum(axis=self.axis), self.axis)
        else:
            return x.sum(axis=self.axis)

class Unsqueeze(Layer):
    name = 'unsqueeze'
    
    def __init__(self, dim):
        self.dim = dim
    
    def forward(self, x):
        return np.expand_dims(np.asarray(x), self.dim)


class Mul(Layer):
    name = 'mul'

    def __init__(self): pass

    def forward(self, x):
        x[1] *= x[0]
        return x[1]


class Const(Layer):
    name = 'const'

    def __init__(self, v): self.v = v

    def forward(self, x): return self.v


class Return(Layer):
    name = 'return'

    def __init__(self): pass

    def forward(self, x):
        return x
    

class BatchNorm(Layer):
    name = 'batchnorm'

    def __init__(self, c):
        self.c = c
        self.k = np.zeros(c, dtype=np.float32)
        self.b = np.zeros(c, dtype=np.float32)
        self.m = np.zeros(c, dtype=np.float32)
        self.v = np.zeros(c, dtype=np.float32)

    def forward(self, x):
        x = x * self.kv_inv
        x += self.kmv_inv_b
        return x
        #return self.kv_inv*x + self.kmv_inv_b

    def load(self, buf):
        c = self.c
        self.k[:] = buf[0*c:1*c]
        self.b[:] = buf[1*c:2*c]
        self.m[:] = buf[2*c:3*c]
        self.v[:] = buf[3*c:4*c]

        self.v_inv = 1/np.sqrt(self.v + 1e-5)
        self.kmv_inv_b = -self.k*self.m*self.v_inv + self.b
        self.kv_inv = self.k*self.v_inv
        self.kmv_inv_b.shape = self.kv_inv.shape = (1,-1,1,1)
        return self.c * 4

class LogSoftmax(Layer):
    name = 'logsoftmax'

    def __init__(self, axis=-1):
        self.axis = axis
        
    def forward(self, x):
        y = x - np.max(x, axis=self.axis, keepdims=True)
        eX = np.sum(np.exp(y), axis=self.axis, keepdims=True)
        y -= np.log(eX); return y

class Shape(Layer):
    name = 'shape'

    def forward(self, x): return x.shape

class Gather(Layer):
    name = 'gather'
    def __init__(self, idx): self.idx = idx

    def forward(self, x): return x[self.idx]

class Reshape(Layer):
    name = 'reshape'

    def forward(self, x): 
        return x[0].reshape(x[1].tolist())

class Transpose(Layer):
    name = 'transpose'
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x): return x.transpose(self.axis)

layer_map = {'dense': Dense, 'conv': Conv2d, 'relu': ReLU, 
             'leakyrelu': LeakyReLU, 'batchnorm': BatchNorm,
             'flatten': Flatten, 'sigmoid': Sigmoid, 'softmax': Softmax, 
             'maxpool': Maxpool, 'avgpool': Avgpool, 'const': Const,
             'upsample': UpSample, 'concat': Concatenate, 'add': Add, 
             'mul': Mul, 'gap': GlobalAveragePool, 'pow':Pow,
             'reducesum':ReduceSum, 'div':Div, 'unsqueeze':Unsqueeze, 
             'shape': Shape, 'gather':Gather, 'reshape':Reshape,
             'transpose':Transpose, 'logsoftmax':LogSoftmax, 'return':Return}

if __name__ == "__main__":
    pass



import networkx as nx
from matplotlib import pyplot as plt

# blue, orange',green,red, purple, brown, pink, gray, olive, cyan
colors = {'conv': 'blue', 'relu': 'orange', 'maxp': 'purple', 'upsa': 'blue', 'dense': 'blue', 'pool': 'olive', 'up': 'orange',
          'conc': 'pink', 'sigm': 'olive', 'batc': 'cyan', 'add': 'purple', 'gap': 'blue', 'flatten': 'pink', 'batchnorm': 'cyan'}

size = {'conv': 300, 'relu': 300, 'maxp': 300, 'upsa': 300, 'dense': 300, 'pool': 300, 'up': 300,
        'conc': 300, 'sigm': 300, 'batc': 300, 'add': 300, 'gap': 300, 'flatten': 300, 'batchnorm': 300}
# ‘so^>v<dph8’.black
shape = {'conv': '^', 'relu': 'p', 'maxp': 'h', 'upsa': 'o', 'dense': 'o', 'pool': 'o', 'up': 'o',
         'conc': '8', 'sigm': 'o', 'batc': '>', 'add': 'o', 'gap': 'o', 'flatten': 'o', 'batchnorm': 0}


def plot_net(body, flow, text=True, nots_colors=colors, nodes_size=size, nodes_shape=shape):
    # read the input/output layer's name(input or input1)
    str_input, str_output = flow[0][0], flow[-1][-1]
    # print(str_input, str_output)
    G = nx.DiGraph()
    dic = {str_input: str_input, 'x': 'x', 'y': 'y'}
    G.add_node(str_input)

    for i in flow:  # add direct edge to the graph
        dic[i[2]] = i[1][-1]
        G.add_node(dic[i[2]])
        # multi-input
        if isinstance(i[0], list):
            for j in i[0]:
                G.add_edge(dic[j], i[1][0])
        else:
            G.add_edge(dic[i[0]], i[1][0])
        # multi-node
        if len(i[1]) > 1:
            for j in range(len(i[1])-1):
                G.add_edge(i[1][j], i[1][j+1])
        if i[2] == str_output:
            G.add_node(str_output)
            G.add_edge(i[1][-1], str_output)

    # positions for all nodes
    pos = nx.kamada_kawai_layout(G)

    # the node's default color is black
    # nx.draw_networkx_nodes(G, pos,  node_color='black')
    # draw the nods using diffenren colors
    for i in nots_colors.keys():
        nx.draw_networkx_nodes(G, pos, nodelist=[j for j in G if i in j],
            node_size=nodes_size[i], node_color=nots_colors[i], node_shape=nodes_shape[i])
    # draw input and outpot layer in red
    nx.draw_networkx_nodes(
        G, pos, nodelist=[str_input, str_output], node_color='red')

    edges = nx.draw_networkx_edges(
        G, pos,  arrowstyle='->', arrowsize=10)

    nx.draw_networkx_labels(G, pos, font_size=8,
        labels={i: i.split('_')[0] for i in G}, font_family='sans-serif')
    plt.axis('off')

    if not text:
        return plt
    # add the para of layer under the nodes
    for i in body:
        x, y = pos[i[0]][0], pos[i[0]][1]-0.08
        if str(i[2]) != 'None':
            plt.text(x, y, s=str(i[2]), fontsize=6,
                     horizontalalignment='center')
    return plt



import numpy as np
np.asnumpy = np.asarray
from time import time

def pad(img, shp, mode='constant', constant_values=0):
    if shp[2][0]==shp[2][1]==shp[3][0]==shp[3][1]==0: return img
    if mode != 'constant': return np.pad(img, shp, mode)
    (n, c, h, w), (mn, mc, mh, mw) = img.shape, shp
    newimg = np.zeros((n, c, h+mh[0]*2, w+mw[0]*2), dtype=img.dtype)
    newimg[:,:,mh[0]:h+mh[0],mw[0]:w+mw[0]] = img
    return newimg

def conv(img, core, group=1, mar=(1, 1), stride=(1, 1), dilation=(1, 1), mode='constant'):
    (strh, strw), (dh, dw) = stride, dilation
    (n, c, h, w), (ni, ci, hi, wi)  = core.shape, img.shape
    cimg_w = c * h * w * group
    cimg_h, i = (hi//strh)*(wi//strw), 0
    shp = ((0, 0), (0, 0), (mar[0],)*2, (mar[1],)*2)
    img = pad(img, shp, mode, constant_values=0)
    img = img.transpose((1,0,2,3)) # nchw -> cnhw
    nh = (hi + sum(shp[2]) - (h-1)*dh-1 + strh)//strh
    nw = (wi + sum(shp[3]) - (w-1)*dw-1 + strw)//strw
    nsh, nsw = nh * strh, nw * strw
    col_img = np.zeros((ci, w*h,  ni, nh, nw), img.dtype) #(h*w, c, N, H, W)
    for r in range(0, h*dh, dh):
        for c in range(0, w*dw, dw):
            col_img[:,i], i = img[:,:,0+r:nsh+r:strh, 0+c:nsw+c:strw], i+1
    col_core = core.reshape((group, core.shape[0]//group, -1))
    col_img.shape = (group, cimg_w//group, -1)
    rst = [i.dot(j) for i, j in zip(col_core, col_img)]
    rst = rst[0] if group==1 else np.concatenate(rst)
    return rst.reshape((n, ni, nh, nw)).transpose(1, 0, 2, 3)

def pool(img, f, core=(2, 2), mar=None, stride=(2, 2),  const=0):
    (n, c, h, w), (ch, cw), (strh, strw) = img.shape, core, stride
    shp = ((0, 0), (0, 0), (mar[0],)*2, (mar[1],)*2)
    img = pad(img, shp, 'constant', constant_values=0)
    (imn, ic, ih, iw), imgs = img.shape, []
    dh = (h + sum(shp[2]) - core[0] + strh)//strh
    dw = (w + sum(shp[3]) - core[1] + strw)//strw
    nsh, nsw = dh * strh, dw * strw
    buf = np.zeros(img.shape[:2]+(dh, dw), np.float32)
    if const != 0: buf[:] = const
    for r in range(0, ch, 1):
        for c in range(0, cw, 1):
            f(img[:,:,r:nsh+r:strh,c:nsw+c:strw], buf, out=buf)
    return buf

def maxpool(i, c=(2, 2), mar=(0,0), s=(2, 2)):
    return pool(i, np.maximum, c, mar, s, -1e4)

def avgpool(i, c=(2, 2), mar=(0,0), s=(2, 2)):
    rst = pool(i, np.add, c, mar, s, 0)
    rst /= c[0] * c[1]
    return rst

def make_upmat(k):
    xs = np.linspace(0.5/k, 1-0.5/k, k*1, dtype=np.float32)
    rs, cs = xs[:,None], xs[None,:]
    klt = ((1-cs)*(1-rs)).reshape((1,-1))
    krt = (cs * (1-rs)).reshape((1,-1))
    klb = ((1-cs) * rs).reshape((1,-1))
    krb = (cs * rs).reshape((1,-1))
    return np.vstack([klt, krt, klb, krb])
    
def upsample_blinear(img, k, matbuf={}):    
    n, c, h, w = img.shape
    img = (img[:,:,:1,:], img, img[:,:,-1:,:])
    img = np.concatenate(img, axis=2)
    img = (img[:,:,:,:1], img, img[:,:,:,-1:])
    img = np.concatenate(img, axis=3)
    if not k in matbuf: matbuf[k] = make_upmat(k)    
    imgs = [img[:,:,:-1,:-1], img[:,:,:-1,1:],
            img[:,:,1:,:-1], img[:,:,1:,1:]]
    imgs = [i[:,:,:,:,None] for i in imgs]
    rst = np.concatenate(imgs, axis=-1)
    rst = np.dot(rst.reshape((-1,4)), matbuf[k])
    rst = rst.reshape((-1, w+1, k, k))
    rst = rst.transpose((0,2,1,3))
    rst = rst.reshape((n,c,(h+1)*k, (w+1)*k))
    return rst[:,:,k//2:-k//2,k//2:-k//2]

def upsample_nearest(img, k):
    n, c, h, w = img.shape
    rst = np.zeros((n, c, h, k, w, k), dtype=np.float32)
    trst = rst.transpose((0,1,2,4,3,5))
    trst[:] = img[:,:,:,:,None,None]
    return rst.reshape((n, c, h*k, w*k))

def upsample(img, k, mode):
    if mode=='nearest': return upsample_nearest(img, k)
    if mode=='linear': return upsample_blinear(img, k)
    
# ===== below is some image process function =====
import math, itertools

def conv_auto(img, core, mode='reflect', keeptp=True):
    shp, dim, (h, w) = img.shape, img.ndim, core.shape
    img = np.pad(img, ((h//2,h//2),(w//2,w//2),(0,0))[:dim], mode=mode)
    rst, buf = np.zeros((2,) + shp, dtype=np.float32)
    for r,c in np.mgrid[:h,:w].reshape(2,-1).T:
        buf[:] = img[r:r+shp[0],c:c+shp[1]]
        buf *= core[r,c]; rst += buf
    return rst.astype(img.dtype) if keeptp else rst

def conv_rc(img, core_r, core_c, mode='reflect'):
    return conv_auto(conv_auto(img, core_r, keeptp=False), core_c)
    
def uniform_filter(img, size=3, mode='reflect'):
    core = np.ones(size, dtype=np.float32)/size
    return conv_rc(img, core[None,:], core[:,None], mode)
    
def gaussian_filter(img, sig=2, mode='reflect'):
    x = np.arange(-int(sig*2.5+0.5), int(sig*2.5+0.5)+1)
    core = np.exp(-x**2/2/sig**2)/sig/(2*np.pi)**0.5
    return conv_rc(img, core[None,:], core[:,None], mode)

def make_slice(l, w, mar):
    r = np.linspace(0, l-w, math.ceil((l-mar)/(w-mar)))
    return [slice(i, i+w) for i in r.astype(int).tolist()]

def grid_slice(H, W, w, h, mar):
    a, b = make_slice(H, w, mar), make_slice(W, h, mar)
    return list(itertools.product(a, b))

def resize(img, size):
    d, (h, w) = img.ndim, img.shape[:2]
    kh, kw = size[0]/h, size[1]/w
    slicer = -0.5+0.5/kh, h-0.5-0.5/kh, size[0]
    rs = np.linspace(*slicer, dtype=np.float32)
    slicec = -0.5+0.5/kw, w-0.5-0.5/kw, size[1]
    cs = np.linspace(*slicec, dtype=np.float32)
    np.clip(rs, 0, h-1, out=rs)
    np.clip(cs, 0, w-1, out=cs)
    ra = np.floor(np.clip(rs, 0, h-1.5))
    ca = np.floor(np.clip(cs, 0, w-1.5))
    ra, ca = ra.astype(int), ca.astype(int)
    rs -= ra; cs -= ca; rb = ra+1; cb = ca+1;
    rs.shape, cs.shape = (-1,1,1)[:d], (1,-1,1)[:d]
    buf = img[:,ca]*(1-cs) + img[:,cb]*cs
    return buf[ra,:]*(1-rs) + buf[rb,:]*rs

def mapcoord(img, rs, cs, keeptp=True):
    d, (h, w) = img.ndim, img.shape[:2]
    np.clip(rs, 0, h-1, out=rs)
    np.clip(cs, 0, w-1, out=cs)
    ra = np.floor(np.clip(rs, 0, h-1.5))
    ca = np.floor(np.clip(cs, 0, w-1.5))
    ra, ca = ra.astype(int), ca.astype(int)
    rs -= ra; cs -= ca; rb = ra+1; cb = ca+1;
    if d==3: rs, cs = rs[:,:,None], cs[:,:,None]
    buf = img[ra,ca]*((1-cs) * (1-rs))
    buf += img[rb,cb] * (cs * rs)
    buf += img[ra,cb] * ((1-rs) * cs)
    buf += img[rb,ca] * ((1-cs) * rs)
    return buf.astype(img.dtype) if keeptp else buf

# sample：缩放，数字代表比例，tuple代表尺寸
# glob：感受野，当缩放后尺寸不足window时，将图像调整为glob整数倍
# window：瓦片尺寸，当缩放后图像大于window，则进行分块处理
# margin：瓦片之间的重叠区域，整数代表宽度，小数代表占瓦片的比例
def tile(sample=1, glob=1, window=1024, margin=0.1, astype='float32'):
    def wrapf(f):
        def wrap(*p, **key):
            (h, w), img = p[0].shape[:2], p[0]
            img = np.asarray(img, dtype=astype)
            tps = {'sample', 'window', 'glob'}
            ftp = fp, tp = {}, {}
            for i in key:
                ftp[i in tps][i] = key[i]
            ssz = tp.get('sample', sample)
            wsz = wsh = wsw = tp.get('window', window)
            gsz = tp.get('glob', glob)
            mar = tp.get('margin', margin)
            info = tp.get('progress', print)
            if isinstance(ssz, tuple): ssz = list(ssz)
            else: ssz = [int(h*ssz), int(w*ssz)]
            # 如果尺寸不足瓦片尺寸，则连同瓦片缩放到glob整数倍
            if wsh>ssz[0]: wsh = ssz[0] = ssz[0]//gsz*gsz
            if wsw>ssz[1]: wsw = ssz[1] = ssz[1]//gsz*gsz
            if ssz!=[h, w]: img = resize(img, ssz)
            if isinstance(mar, float): mar = int(wsz*mar)
            rcs = grid_slice(*ssz, wsh, wsw, mar)
            if len(rcs)>1: info(1, len(rcs))
            rst = f(img[rcs[0]], *p[1:], **fp)
            if len(rcs)==1 and ssz!=[h, w]:
                rst = resize(rst, (h,w))
            if len(rcs)==1: return np.asnumpy(rst)
            outshp = img.shape[:2] + rst.shape[2:]
            weights = np.zeros(rst.shape[:2], dtype='uint16')
            if rst.ndim==3: weights = weights[:,:,None]
            weights += mar + 1
            for i in range(mar, 0, -1):
                weights[i-1,:] = weights[-i,:] = i
                weights[:,i-1] = weights[:,-i] = i
            rst *= weights
            buf = np.zeros(outshp, dtype=rst.dtype)
            count = np.zeros(outshp[:2], dtype='uint16')
            if rst.ndim==3: count = count[:,:,None]
            buf[rcs[0]] = rst; count[rcs[0]] += weights
            for i in range(1, len(rcs)):
                info(i+1, len(rcs))
                rst = f(img[rcs[i]], *p[1:], **fp)
                rst *= weights
                buf[rcs[i]] += rst; count[rcs[i]] += weights
            np.divide(buf, count, out=buf, casting='unsafe')
            if ssz!=(h, w): buf = resize(buf, (h,w))
            return np.asnumpy(buf)
        return wrap
    return wrapf

if __name__ == '__main__':
    '''
    img = np.zeros((1, 64, 512, 512), dtype=np.float32)
    core = np.zeros((32, 64, 3, 3), dtype=np.float32)
    conv(img, core)
    '''
    import numpy as np
    a = np.arange(12).reshape(2,2,3)



from setuptools import setup

descr = """A powerful light-weight inference framework for CNN.
The aim of planer is to provide efficient and adaptable inference environment for CNN model.
Also in order to enlarge the application scope,
we support ONNX format, which enables the converting of trained model within various DL frameworks.
"""

if __name__ == '__main__':
    setup(name='planer',
        version='0.19',
        url='https://github.com/Image-Py/planer',
        description='Powerful Light Artificial NEuRon',
        long_description=descr,
        author='Y.Dong, YXDragon',
        author_email='yxdragon@imagepy.org',
        license='BSD 3-clause',
        packages=['planer'],
        package_data={},
        install_requires=[
            'numpy'
        ],
    )



from .layer import layer_map as key
from time import time
from .util import np

class Net:
    def __init__(self):
        self.body, self.flow = [], []
        self.life, self.timer = {}, {}

    def load_json(self, body, flow, debug=False):
        self.body, self.flow, self.life = [], [], {}
        for i in body:
            para = i[2] or []
            if debug: print(i)
            self.body.append((i[0], key[i[1]](*para)))
        for i in range(len(flow)):
            keys = flow[i][0]
            if isinstance(keys, str): keys = [keys]
            for j in keys: self.life[j] = i
            
        self.layer, self.flow = body, flow

    def info(self, obj):
        if isinstance(obj, list):
            return [self.info(i) for i in obj]
        if hasattr(obj, 'shape'): return obj.shape
        return obj

    def forward(self, x, debug=False):
        dic = dict(self.body)
        rst = {self.flow[0][0]: x, 'None': None}
        for i in range(len(self.flow)):
            x, ls, y = self.flow[i]
            if not isinstance(ls, list): ls = [ls]
            for l in ls:
                out = x if l == ls[0] else y
                if not isinstance(out, str):
                    p = [rst[i] for i in out]
                else: p = rst[out]
                xs = x if isinstance(x, list) else [x]
                for k in xs: # release wasted obj
                    if self.life[k]<=i: del rst[k]
                obj = dic[l]
                start = time()
                if debug: print(obj.name, ':', obj.para())
                if debug: print('--> ', out, ':', self.info(p))
                rst[y] = obj(p)
                if debug: print('<-- ',  y, ':', self.info(rst[y]))
                cost = time()-start
                if not obj.name in self.timer:
                    self.timer[obj.name] = 0
                self.timer[obj.name] += cost
        return rst[y][0] if len(rst[y])==1 else rst[y]

    def layer2code(self, style='list'):
        body = []
        if style == 'list':
            body = ['self.layer = [']
            for i in self.body:
                body.append('\t("%s", %s, %s),' % (i[0],
                    i[1].__class__.__name__, i[1].para()))
            body.append(']')
        if style == 'self':
            body = []
            for i in self.body:
                body.append('self.%s = %s%s' % (i[0],
                    i[1].__class__.__name__, i[1].para() or ()))
        return '\n'.join(body)

    def layer2json(self):
        body = []
        invk = dict(zip(key.values(), key.keys()))
        for i in self.body:
            body.append((i[0], invk[i[1].__class__], i[1].para()))
        return body

    def flw2code(self, style='list'):
        body = []
        if style == 'list':
            for x, ls, y in self.cmds:
                for l in ls:
                    out = x if l == ls[0] else y
                    if isinstance(out, list):
                        out = str(out).replace("'", '')
                    body.append("%s = dic['%s'](%s)" % (y, l, out))
                body.append('')
        if style == 'self':
            for x, ls, y in self.cmds:
                for l in ls:
                    out = x if l == ls[0] else y
                    if isinstance(out, list):
                        out = str(out).replace("'", '')
                    body.append('%s = self.%s(%s)' % (y, l, out))
                body.append('')
        return '\n'.join(body)

    def load_weights(self, data):
        s = 0
        for i in self.body:
            s += i[1].load(data[s:])


    def show(self, info=True):
        from .plot import plot_net
        plot_net(self.layer, self.flow, info).show()

    def __call__(self, x, debug=False):
        return self.forward(x, debug)


if __name__ == '__main__':
    pass



import json, re, os
import numpy as np
from .net import Net
from time import time
import json, zipfile
from io import BytesIO

def read_net(path, debug=False):
    net = Net()
    if os.path.exists(path+'.pla'):
        with zipfile.ZipFile(path+'.pla') as f:
            path = os.path.split(path)[1]
            body = json.loads(f.read(path+'.json'))
            lay, flw = body['layers'], body['flow']
            buf = BytesIO(f.read(path+'.npy'))
            weights = np.load(buf)
    elif os.path.exists(path+'.json'):
        with open(path+'.json') as f:
            body = json.load(f)
            lay, flw = body['layers'], body['flow']
        weights = np.load(path+'.npy')
    else: return print('model %s not found!'%path)
    net.load_json(lay, flw, debug)
    net.load_weights(weights)
    return net

def onnx2planer(path):
    import onnx, onnx.numpy_helper
    graph = onnx.load(path).graph
    layers, weights, flows, const, values = [], [], [], {}, {}
    for i in graph.initializer: values[i.name] = onnx.numpy_helper.to_array(i)
    for i in graph.node:
        # print(i.name)
        has = [i for i in i.input if i in values]
        no = [i for i in i.input if not i in values]
        if len(no)==1: no = no[0]
        if no != []:
            flows.append([no, [i.name], i.output[0]])
            weights.extend([values[i] for i in has])
        if i.op_type == 'BatchNormalization':
            layers.append([i.name, 'batchnorm', [weights[-1].shape[0]]])
        elif i.op_type == 'Conv':
            attr, w = i.attribute, weights[-2].shape
            g, d, p, s = attr[1].i, attr[0].ints[0], attr[3].ints[0], attr[4].ints[0]
            layers.append([i.name, 'conv', [w[1], w[0], w[2], g, s, d, p]])
        elif i.op_type == 'Gemm':
            layers.append([i.name, 'dense', list(weights[-2].shape[::-1])])
        elif i.op_type == 'MaxPool':
            w, m, s = i.attribute[0].ints, i.attribute[1].ints, i.attribute[2].ints
            layers.append([i.name, 'maxpool', [list(i) for i in [w, m[2:], s]]])
        elif i.op_type == 'GlobalAveragePool':
            layers.append([i.name, 'gap', None])
        elif i.op_type == 'Upsample':
            mode, k = i.attribute[0].s.decode('utf-8'), weights.pop(-1)
            layers.append([i.name, 'upsample', [int(k[-1]), mode]])
        elif i.op_type == 'Flatten':
            layers.append([i.name, 'flatten', None])
        elif i.op_type == 'Unsqueeze':
            layers.append([i.name, 'unsqueeze', [i.attribute[0].ints[0]]])
        elif i.op_type == 'Relu':
            layers.append([i.name, 'relu', None])
        elif i.op_type == 'LeakyRelu':
            alpha = i.attribute[0].f
            layers.append([i.name, 'leakyrelu', [alpha]])
        elif i.op_type == 'Add':
            layers.append([i.name, 'add', None])
        elif i.op_type == 'Div':
            layers.append([i.name, 'div', None])
        elif i.op_type == 'Constant':
            buf = i.attribute[0].t.raw_data
            tp = i.attribute[0].t.data_type
            tp, fmt = [(int, 'int64'), (float, 'float32')][tp==1]
            const[i.output[0]] = tp(np.frombuffer(buf, fmt))
        elif i.op_type == 'Pow':
            flows[-1][0], k = flows[-1][0]
            layers.append([i.name, 'pow', [const[k]]])
        elif i.op_type == 'ReduceSum':
            axis, keep = i.attribute[0].ints[0], i.attribute[1].i
            layers.append([i.name, 'reducesum', [axis, keep]])
        elif i.op_type == 'Concat':
            layers.append([i.name, 'concat', [i.attribute[0].i]])
        elif i.op_type == 'Pad':
            layers.append([i.name, 'identity', None])
        elif i.op_type == 'Sigmoid':
            layers.append([i.name, 'sigmoid', None])
        elif i.op_type == 'AveragePool':
            for attr in i.attribute:
                if 'stride' in attr.name: s = attr.ints[0]
                if 'kernel' in attr.name: w = attr.ints[0]
            layers.append([i.name, 'avgpool', [w, s]])
        elif i.op_type == 'Shape':
            layers.append([i.name, 'shape', None])
        elif i.op_type == 'Gather':
            flows[-1][0], k = flows[-1][0]
            layers.append([i.name, 'gather', [const[k]]])
        elif i.op_type == 'Mul':
            layers.append([i.name, 'mul', None])
        elif i.op_type == 'Reshape':
            layers.append([i.name, 'reshape', None])
        elif i.op_type == 'Transpose':
            layers.append([i.name, 'transpose', [list(i.attribute[0].ints)]])
        elif i.op_type == 'LogSoftmax':
            layers.append([i.name, 'logsoftmax', [i.attribute[0].i]])
        else:
            print('lost layer:', i.name)
            return i
    layers.append(['return', 'return', None])
    flows.append([[i.name for i in graph.output], ['return'], 'plrst'])
    weights = np.hstack([i.ravel() for i in weights])

    np.save(path.replace('onnx', 'npy'), weights)
    with open(path.replace('onnx', 'json'), 'w') as f:
        json.dump({'layers':layers, 'flow':flows}, f)

    with zipfile.ZipFile(path.replace('onnx', 'pla'), 'w') as f:
        f.write(path.replace('onnx','json'))
        f.write(path.replace('onnx','npy'))

if __name__ == '__main__':
    a, b = read_onnx('../demo/yolov3-planer-2/yolov3')
